# -*- coding: utf-8 -*-
"""
Created on Mon Jul 11 17:52:02 2016

@author: caselitz
"""

import numpy as np
import matplotlib.pyplot as plt


def get_neighborhood(cell, occ_map_shape):

  n = np.array([]).reshape(0,2)
  
  ### YOUR CODE FOR FINDING THE NEIGHBORHOOD OF cell GOES HERE ###
  
  # Return a nx2 array in the format [[n1_y,n1_x],[n2_y,n2_x],...]
  
  return n


def get_edge_cost(parent, child, occ_map):
  
  edge_cost = 0
  
  ### YOUR CODE FOR CALCULATING THE COST FROM parent TO child GOES HERE
  
  return edge_cost


def get_heuristic(cell, goal):
  
  cost = 0
  
  ### YOUR CODE FOR ESTIMATING THE COST FROM cell TO goal GOES HERE
  
  return cost
  

def main():  
  
  #load the map
  occ_map = np.loadtxt('map.txt')
  
  #show map
  plt.matshow(occ_map, fignum=1, cmap=plt.cm.gray)
  plt.axis([0, occ_map.shape[1]-1, occ_map.shape[0]-1, 0])
  plt.xlabel('x')
  plt.ylabel('y')
  plt.ion()
  
  #cost value for each cell
  cost = np.ones(occ_map.shape)*np.inf
  
  #heuristic that estimates costs to the goal
  heuristic = np.zeros(occ_map.shape)
  
  #cells that have been expanded
  closed_list = np.zeros(occ_map.shape)
   
  #implicitly store path by remembering predecessors
  predecessor_y = np.zeros(occ_map.shape, dtype=np.int)-1
  predecessor_x = np.zeros(occ_map.shape, dtype=np.int)-1
  
  # start and goal position [y,x]
  start = np.array([33,22])
  goal = np.array([15,40])
  
  #plot start and goal
  plt.plot([start[1]],[start[0]],'ro')
  plt.plot([goal[1]],[goal[0]],'go')
  
  #start search at the start
  parent=start
  cost[start[0],start[1]] = 0
  
  
  while not np.array_equal(parent,goal):
    
    #candidate cells for expansion (i.e. not in closed list)  
    open_list = np.where(closed_list==1,np.inf,cost) + heuristic
    
    #find cell with minimum cost in the open list
    y, x = np.unravel_index(open_list.argmin(),open_list.shape)
    
    #break loop if open list is empty
    if np.isinf(open_list[y,x]):
      break
    
    #set parent and put it in closed list
    parent = np.array([y,x])
    closed_list[y,x] = 1;
    
    #get neighborhood of parent  
    neighbors = get_neighborhood(parent,occ_map.shape)
    
    #for each neighbor
    for i in xrange(neighbors.shape[0]):
      
      child = neighbors[i]

      #compute costs
      child_cost = cost[y,x] + get_edge_cost(parent,child,occ_map) 
      
      #check if cost is lower
      if child_cost < cost[child[0],child[1]]:
        #update cost
        cost[child[0],child[1]] = child_cost
        #compute heuristic
        heuristic[child[0],child[1]] = get_heuristic(child,goal)
        #set predecessor
        predecessor_y[child[0],child[1]] = y
        predecessor_x[child[0],child[1]] = x     
  
    #plot parent (uncomment draw() for animation)
    if not (np.array_equal(parent,start) or np.array_equal(parent,goal)):
      plt.plot([parent[1]],[parent[0]],'yo')
      plt.draw()
  
  
  if np.array_equal(parent,goal):
    
    print "found goal: " + str(parent) 
    print "cells expanded: " + str(np.count_nonzero(closed_list))
    print "path cost: " + str(cost[parent[0],parent[1]])
    
    path_length = 0
    
    #rewind the path by from goal to start (at start predecessor is -1)  
    while predecessor_x[parent[0],parent[1]] >= 0:
      
      #predecessor
      predecessor = np.array([predecessor_y[parent[0],parent[1]],
                              predecessor_x[parent[0],parent[1]]])
      
      #plot path (uncomment draw() for animation)
      if not np.array_equal(parent,goal):
        plt.plot([parent[1]],[parent[0]],'bo')  
        plt.draw()
      
      #compute path length
      path_length = path_length + np.linalg.norm(parent-predecessor)
      
      #set parent to predecessor
      parent = predecessor
    
    print "path length: " + str(path_length)
  
  else:
    print "no valid path found"


  #plot cost
  plt.ioff()
  plt.matshow(cost, fignum=2, cmap=plt.cm.gray)
  plt.axis([0, occ_map.shape[1]-1, occ_map.shape[0]-1, 0])
  plt.xlabel('x')
  plt.ylabel('y')
  plt.show()

if __name__ == "__main__":
  main()